# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-24 21:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_countries.fields
import driver27.stats
import swapfield.fields


class Migration(migrations.Migration):

    replaces = [(b'driver27', '0001_initial'), (b'driver27', '0002_auto_20161103_0846'), (b'driver27', '0003_auto_20170204_2308'), (b'driver27', '0004_auto_20170204_2311'), (b'driver27', '0005_result_points'), (b'driver27', '0006_auto_20170529_2209'), (b'driver27', '0007_auto_20170529_2211'), (b'driver27', '0008_auto_20170529_2220'), (b'driver27', '0009_populate_driver_in_seats'), (b'driver27', '0010_seats_contender_null'), (b'driver27', '0011_remove_contender_teams'), (b'driver27', '0012_auto_20170529_2232'), (b'driver27', '0013_remove_seat_contender'), (b'driver27', '0014_auto_20170529_2320'), (b'driver27', '0015_auto_20170602_1447'), (b'driver27', '0016_driver_teams'), (b'driver27', '0017_auto_20170602_1526'), (b'driver27', '0018_populate_competition_team'), (b'driver27', '0019_remove_team_competitions'), (b'driver27', '0020_team_competitions'), (b'driver27', '0021_auto_20170602_1547'), (b'driver27', '0022_auto_20170602_1552'), (b'driver27', '0023_auto_20170602_1552'), (b'driver27', '0024_auto_20170602_1737'), (b'driver27', '0025_auto_20171022_2012'), (b'driver27', '0026_populate_fastest_race'), (b'driver27', '0027_auto_20171022_2021')]


    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Circuit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='circuit')),
                ('city', models.CharField(blank=True, max_length=100, null=True, verbose_name='city')),
                ('country', django_countries.fields.CountryField(max_length=2, verbose_name='country')),
                ('opened_in', models.IntegerField(verbose_name='opened in')),
            ],
            options={
                'ordering': ['name'],
                'verbose_name': 'Circuit',
                'verbose_name_plural': 'Circuits',
            },
        ),
        migrations.CreateModel(
            name='Competition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='competition')),
                ('full_name', models.CharField(max_length=100, unique=True, verbose_name='full name')),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True, verbose_name='country')),
                ('slug', models.SlugField(blank=True, default=None, null=True)),
            ],
            options={
                'ordering': ['name'],
                'verbose_name': 'Competition',
                'verbose_name_plural': 'Competitions',
            },
        ),
        migrations.CreateModel(
            name='CompetitionTeam',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_year', models.IntegerField(blank=True, default=None, null=True)),
                ('until_year', models.IntegerField(blank=True, default=None, null=True)),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Competition', verbose_name='competition')),
            ],
        ),
        migrations.CreateModel(
            name='Driver',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_name', models.CharField(max_length=50, verbose_name='last name')),
                ('first_name', models.CharField(max_length=25, verbose_name='first name')),
                ('year_of_birth', models.IntegerField(blank=True, null=True, verbose_name='year of birth')),
                ('country', django_countries.fields.CountryField(blank=True, max_length=2, null=True, verbose_name='country')),
            ],
            options={
                'ordering': ['last_name', 'first_name'],
                'verbose_name': 'Driver',
                'verbose_name_plural': 'Drivers',
            },
            bases=(models.Model, driver27.stats.AbstractStreakModel),
        ),
        migrations.CreateModel(
            name='GrandPrix',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True, verbose_name='grand prix')),
                ('country', django_countries.fields.CountryField(blank=True, default=None, max_length=2, null=True, verbose_name='country')),
                ('first_held', models.IntegerField(blank=True, null=True, verbose_name='first held')),
                ('competitions', models.ManyToManyField(default=None, related_name='grands_prix', to='driver27.Competition', verbose_name='competitions')),
                ('default_circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='default_to_grands_prix', to='driver27.Circuit', verbose_name='default circuit')),
            ],
            options={
                'ordering': ['name'],
                'verbose_name': 'Grand Prix',
                'verbose_name_plural': 'Grands Prix',
            },
        ),
        migrations.CreateModel(
            name='Race',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('round', models.IntegerField(verbose_name='round')),
                ('date', models.DateField(blank=True, default=None, null=True, verbose_name='date')),
                ('alter_punctuation', models.CharField(blank=True, choices=[('double', 'Double'), ('half', 'Half')], default=None, max_length=6, null=True, verbose_name='alter punctuation')),
                ('circuit', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Circuit', verbose_name='circuit')),
            ],
            options={
                'ordering': ['season', 'round'],
                'verbose_name': 'Race',
                'verbose_name_plural': 'Races',
            },
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('qualifying', swapfield.fields.SwapIntegerField(blank=True, default=None, null=True, unique_for_fields=['race'], verbose_name='qualifying')),
                ('finish', swapfield.fields.SwapIntegerField(blank=True, default=None, null=True, unique_for_fields=['race'], verbose_name='finish')),
                ('retired', models.BooleanField(default=False, verbose_name='retired')),
                ('wildcard', models.BooleanField(default=False, verbose_name='wildcard')),
                ('comment', models.CharField(blank=True, default=None, max_length=250, null=True, verbose_name='comment')),
                ('points', models.IntegerField(blank=True, default=0, null=True, verbose_name='points')),
                ('race', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Race', verbose_name='race')),
            ],
            options={
                'ordering': ['race__season', 'race__round', 'finish', 'qualifying'],
                'verbose_name': 'Result',
                'verbose_name_plural': 'Results',
            },
        ),
        migrations.CreateModel(
            name='Season',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('year', models.IntegerField(verbose_name='year')),
                ('rounds', models.IntegerField(blank=True, default=None, null=True, verbose_name='rounds')),
                ('punctuation', models.CharField(default=None, max_length=20, null=True, verbose_name='punctuation')),
                ('competition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons', to='driver27.Competition', verbose_name='competition')),
            ],
            options={
                'ordering': ['year', 'competition__name'],
                'verbose_name': 'Season',
                'verbose_name_plural': 'Seasons',
            },
        ),
        migrations.CreateModel(
            name='Seat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('driver', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Driver', verbose_name='driver')),
            ],
            options={
                'ordering': ['driver__last_name', 'team'],
                'verbose_name': 'Seat',
                'verbose_name_plural': 'Seats',
            },
        ),
        migrations.CreateModel(
            name='SeatPeriod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_year', models.IntegerField(blank=True, default=None, null=True)),
                ('until_year', models.IntegerField(blank=True, default=None, null=True)),
                ('seat', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Seat')),
            ],
        ),
        migrations.CreateModel(
            name='Team',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=75, unique=True, verbose_name='team')),
                ('full_name', models.CharField(max_length=200, unique=True, verbose_name='full name')),
                ('country', django_countries.fields.CountryField(blank=True, max_length=2, null=True, verbose_name='country')),
                ('competitions', models.ManyToManyField(related_name='teams', through='driver27.CompetitionTeam', to='driver27.Competition', verbose_name='competitions')),
            ],
            options={
                'ordering': ['name'],
                'verbose_name': 'Team',
                'verbose_name_plural': 'Teams',
            },
            bases=(models.Model, driver27.stats.AbstractStreakModel),
        ),
        migrations.CreateModel(
            name='TeamSeason',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sponsor_name', models.CharField(blank=True, default=None, max_length=75, null=True, verbose_name='sponsor name')),
                ('season', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='teams_season', to='driver27.Season', verbose_name='season')),
                ('team', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seasons_team', to='driver27.Team', verbose_name='team')),
            ],
            options={
                'verbose_name': 'Team Season',
                'verbose_name_plural': 'Teams Season',
            },
            bases=(models.Model, driver27.stats.AbstractStreakModel, driver27.stats.SeasonStatsModel),
        ),
        migrations.AddField(
            model_name='seat',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='seats', to='driver27.Team', verbose_name='team'),
        ),
        migrations.AddField(
            model_name='result',
            name='seat',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='driver27.Seat', verbose_name='seat'),
        ),
        migrations.AddField(
            model_name='race',
            name='fastest_car',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fastest_in', to='driver27.Seat'),
        ),
        migrations.AddField(
            model_name='race',
            name='grand_prix',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.GrandPrix', verbose_name='grand prix'),
        ),
        migrations.AddField(
            model_name='race',
            name='season',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='races', to='driver27.Season', verbose_name='season'),
        ),
        migrations.AddField(
            model_name='driver',
            name='teams',
            field=models.ManyToManyField(related_name='drivers', through='driver27.Seat', to='driver27.Team', verbose_name='teams'),
        ),
        migrations.AddField(
            model_name='competitionteam',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='driver27.Team', verbose_name='team'),
        ),
        migrations.CreateModel(
            name='SeatsSeason',
            fields=[
            ],
            options={
                'verbose_name': 'Seats by season',
                'proxy': True,
                'verbose_name_plural': 'Seats by season',
                'indexes': [],
            },
            bases=('driver27.season',),
        ),
        migrations.AlterUniqueTogether(
            name='teamseason',
            unique_together=set([('season', 'team')]),
        ),
        migrations.AlterUniqueTogether(
            name='seat',
            unique_together=set([('team', 'driver')]),
        ),
        migrations.AlterUniqueTogether(
            name='season',
            unique_together=set([('year', 'competition')]),
        ),
        migrations.AlterUniqueTogether(
            name='result',
            unique_together=set([('race', 'seat')]),
        ),
        migrations.AlterUniqueTogether(
            name='race',
            unique_together=set([('season', 'round')]),
        ),
        migrations.AlterUniqueTogether(
            name='driver',
            unique_together=set([('last_name', 'first_name')]),
        ),
        migrations.AlterUniqueTogether(
            name='competitionteam',
            unique_together=set([('team', 'competition', 'from_year', 'until_year')]),
        ),
    ]



# -*- coding: utf-8 -*-
from django.test import TestCase
from .common import retro_encode, CommonCompetitionTestCase
from ..models import Competition
from slugify import slugify

class CompetitionTestCase(TestCase, CommonCompetitionTestCase):
    def test_competition_unicode(self):
        competition = self.get_test_competition()
        self.assertEquals(competition.slug, slugify(competition.name))
        self.assertEqual(str(competition), retro_encode(competition.name))


# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import CommonSeatTestCase, CommonSeasonTestCase
from ..models import ContenderSeason


class ContenderSeasonTestCase(TestCase, CommonSeatTestCase, CommonSeasonTestCase):
    def test_contender_season(self):
        driver = None
        season = None
        self.assertRaises(AttributeError, ContenderSeason,
                          **{'driver': driver, 'season': season})
        seat = self.get_test_seat()
        competition = self.get_test_competition()
        season = self.get_test_season(competition=competition)
        driver = seat.driver
        self.assertTrue(ContenderSeason(driver=driver, season=season))


from django.conf.urls import url, include

from .. import views
from ..api import urls_api
from ..common import DRIVER27_NAMESPACE


def driver27_urls():
    return [
        url(r'^$', views.global_view, name='view'),
        url(r'^ajax/', include('driver27.urls.ajax', namespace='dr27-ajax')),
        url(r'^global/', include('driver27.urls.global', namespace='global')),
        url(r'^api/auth/', include('rest_framework.urls')),
        url(r'^api/', include(urls_api)),
        url(r'^(?P<competition_slug>[-\w\d]+)/(?P<year>\d+)/', include('driver27.urls.season', namespace='season')),
        url(r'^(?P<competition_slug>[-\w\d]+)/', include('driver27.urls.competition', namespace='competition')),
    ]


urlpatterns = [
    url(r'^', include(driver27_urls(), namespace=DRIVER27_NAMESPACE)),
]



from django.db.models.fields import BLANK_CHOICE_DASH
from .common import *
from ..punctuation import get_punctuation_label_dict
from ..models import Season


class SeasonAdminForm(AlwaysChangedModelForm):
    def __init__(self, *args, **kwargs):
        super(SeasonAdminForm, self).__init__(*args, **kwargs)
        punctuation_choices = get_punctuation_label_dict()
        self.fields['punctuation'] = forms.ChoiceField(choices=BLANK_CHOICE_DASH + list(punctuation_choices),
                                                       initial=None)
        self.fields['rounds'].help_text = 'If \'rounds\' are not empty, only the best (rounds) results will be ' \
                                          'counted for driver points'

    class Meta:
        model = Season
        fields = ('year', 'competition', 'rounds', 'punctuation')


class RaceAdminForm(forms.ModelForm):

    def __init__(self, instance, *args, **kwargs):
        super(RaceAdminForm, self).__init__(instance=instance, *args, **kwargs)
        if instance.pk:
            fastest_choices = list(instance.seats.values_list('pk', flat=True))
            qs = self.fields['fastest_car'].queryset.filter(pk__in=fastest_choices)
            self.fields['fastest_car'].queryset = qs

    class Meta:
        widgets = {
            'grand_prix': GrandPrixWidget,
        }



from operator import le, ge, eq, lt, gt
import re
from django.db.models import F


def build_dict(seq, key):
    return dict((d[key], dict(d, index=index)) for (index, d) in enumerate(seq))


class Streak(object):
    results = None
    max_streak = False

    def __init__(self, results, unique_by_race=False, max_streak=False):
        self.results = results
        self.max_streak = max_streak
        self.unique_by_race = unique_by_race

    @staticmethod
    def builtin_function_dict():
        return {'lte': le,
                'gte': ge,
                'eq': eq,
                'exact': eq,
                'lt': lt,
                'gt': gt
                }

    @staticmethod
    def get_builtin_function(builtin_key):
        return Streak.builtin_function_dict().get(builtin_key, None)

    @staticmethod
    def validate_syntax_filter(current_filter):
        if not re.match(r'^(\w)+$', current_filter):
            raise Exception('Invalid streak filter syntax')

    @staticmethod
    def split_filter(curr_filter):
        split_char = '__'
        return curr_filter.split(split_char)

    @staticmethod
    def get_comparison_position(filter_list):
        return len(filter_list) - 1

    @classmethod
    def get_comparison_item(cls, filter_list):
        comparison_position = cls.get_comparison_position(filter_list)
        return filter_list[comparison_position]

    @classmethod
    def exists_comparison(cls, filter_list):
        comparison_position = cls.get_comparison_position(filter_list)
        return filter_list[comparison_position] in Streak.builtin_function_dict()

    @classmethod
    def exec_comparison(cls, result, filter_list, filter_value):
        comparison_key = 'eq'
        if cls.exists_comparison(filter_list):
            comparison_key = cls.get_comparison_item(filter_list)
            filter_list.pop(cls.get_comparison_position(filter_list))
        filter_key = '.'.join(filter_list)
        result_attr = getattr(result, filter_key)
        if isinstance(filter_value, F):
            filter_value = getattr(result, filter_value.name)
        builtin_function = cls.get_builtin_function(comparison_key)
        return builtin_function(result_attr, filter_value) if builtin_function else False

    @classmethod
    def checked_filters(cls, result, filters):
        passed_filter = True  # until is a false condition, filter is passed
        for curr_filter in filters:
            filter_value = filters[curr_filter]
            filter_list = cls.split_filter(curr_filter)
            comparison_result = cls.exec_comparison(result, filter_list, filter_value)
            if not comparison_result:
                passed_filter = False
                break
        return passed_filter

    def get_results_by_race(self):
        results_by_race = []
        race_order = 0
        for result in self.results:
            race_str = 'race_{id}'.format(id=result.race_id)

            race_indexes = build_dict(results_by_race, key="race_id")
            if race_str not in race_indexes:
                results_by_race.append(
                    {
                        'race_id': race_str,
                        'order': race_order,
                        'results': []
                    }
                )

                race_order += 1
                results_by_race[-1]['results'].append(result)
            else:
                race_index = race_indexes[race_str]['index']
                results_by_race[race_index]['results'].append(result)
        return sorted(results_by_race, key=lambda x: x['order'])

    def run_unique_by_race(self, filters):
        results_by_race = self.get_results_by_race()
        count = 0
        max_count = 0
        for race_results in results_by_race:
            is_ok = any([self.checked_filters(result, filters) for result in race_results['results']])
            if not is_ok:
                if self.max_streak:
                    count = 0
                    continue
                else:
                    break
            count += 1
            if self.max_streak and count > max_count:
                max_count = count
        return max_count if self.max_streak else count

    def run(self, filters):
        if self.unique_by_race:
            return self.run_unique_by_race(filters)
        count = 0
        max_count = 0
        for result in self.results:
            is_ok = self.checked_filters(result, filters)

            if not is_ok:
                if self.max_streak:
                    count = 0
                    continue
                else:
                    break
            count += 1

            if self.max_streak and count > max_count:
                max_count = count
        return max_count if self.max_streak else count



from django.conf.urls import url, include

from driver27 import views

urlpatterns = [
    url(r'^$', views.global_view, name='index'),
    url(r'^tpl/$', views.global_tpl, name='tpl'),
    url(r'^driver/(?P<driver_id>\d+)/$', views.driver_profile_view, name='driver-profile'),
    url(r'^driver/', include('driver27.urls.driver')),
    url(r'^team/(?P<team_id>\d+)/$', views.team_profile_view, name='team-profile'),
    url(r'^team/', include('driver27.urls.team')),
]



from django.conf import settings
from django.contrib.admin.sites import AdminSite
from django.test import TestCase, Client, RequestFactory

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse
from django.forms.models import inlineformset_factory
from ..models import Season, Driver, Team, Competition, Circuit, get_tuples_from_results
from ..models import GrandPrix, Race, Seat, TeamSeason, ContenderSeason
from ..admin import SeasonAdmin, SeasonAdminForm, DriverAdmin, TeamAdmin
from ..admin import CompetitionAdmin, CircuitAdmin, GrandPrixAdmin
from ..admin import RaceAdmin, RelatedCompetitionAdmin
from ..admin import RaceInline, TeamSeasonInline
from ..admin.formsets import RaceFormSet
from ..admin.common import AlwaysChangedModelForm
from ..punctuation import get_punctuation_config
from rest_framework.test import APITestCase
from ..common import DRIVER27_NAMESPACE, DRIVER27_API_NAMESPACE

class MockRequest(object):
    pass


class MockSuperUser(object):
    def has_perm(self, perm):
        return True


def get_request():
    request = MockRequest()
    request.user = MockSuperUser()
    return request


def get_fixtures_test():
    # django test uses a new db, while pytest use transaction no_commit in the same settings db
    if hasattr(settings, 'PYTEST_SETTING') and settings.PYTEST_SETTING:
        return None
    else:
        return ['driver27.json', ]


class FixturesTest(TestCase):
    fixtures = get_fixtures_test()


class ViewTest(FixturesTest):
    def setUp(self):
        self.site = AdminSite()
        self.client = Client()
        self.factory = RequestFactory()

    def _GET_request(self, reverse_url, kwargs=None, data=None, code=200):
        # Issue a GET request.
        reverse_url = DRIVER27_NAMESPACE+':'+reverse_url
        the_reverse = reverse(reverse_url, kwargs=kwargs)
        response = self.client.get(the_reverse, data=data)
        # Check that the response is 302 OK.
        self.assertEqual(response.status_code, code)

    # def test_competition_list(self):
    #     self._GET_request('competition:list')

    def _test_competition_view(self):
        kwargs = {'competition_slug': 'f1'}
        self._GET_request('competition:view', kwargs=kwargs)
        self._GET_request('competition:driver-olympic', kwargs=kwargs)
        self._GET_request('competition:driver-comeback', kwargs=kwargs)
        self._GET_request('competition:driver-record-index', kwargs=kwargs)
        self._GET_request('competition:team-olympic', kwargs=kwargs)
        self._GET_request('competition:driver-seasons-rank', kwargs=kwargs)
        self._GET_request('competition:team-seasons-rank', kwargs=kwargs)
        self._GET_request('competition:driver-record-index', kwargs=kwargs)
        self._GET_request('competition:view', kwargs={'competition_slug': 'f19'}, code=404)

    def _test_season_view(self):
        kwargs = {'competition_slug': 'f1', 'year': 2016}
        self._GET_request('season:view', kwargs=kwargs)
        self._GET_request('season:race-list', kwargs=kwargs)
        race_kw = {'race_id': 1}
        race_kw.update(kwargs)
        self._GET_request('season:race-view', kwargs=race_kw)
        self._GET_request('season:driver', kwargs=kwargs)
        self._GET_request('season:driver-olympic', kwargs=kwargs)
        self._GET_request('season:driver-comeback', kwargs=kwargs)
        self._GET_request('season:driver-record-index', kwargs=kwargs)
        self._GET_request('season:team', kwargs=kwargs)
        self._GET_request('season:team-olympic', kwargs=kwargs)
        self._GET_request('season:team-record-index', kwargs=kwargs)
        self._GET_request('season:race-list', kwargs=kwargs)

        kwargs = {'competition_slug': 'f19', 'year': 2006}
        self._GET_request('season:view', kwargs=kwargs, code=404)

    def test_ajax_standing(self, model='driver', default_code=200):
        URL = 'dr27-ajax:standing'
        data = {'model': model}
        self._GET_request(URL, data=data, code=default_code)
        data['competition_slug'] = 'f1'
        self._GET_request(URL, data=data, code=default_code)
        data['year'] = 2016
        self._GET_request(URL, data=data, code=default_code)
        data['olympic'] = True
        self._GET_request(URL, data=data, code=default_code)

    def test_ajax_standing_team(self):
        self.test_ajax_standing(model='team')

    def test_ajax_standing_404_model(self):
        self.test_ajax_standing(model='else', default_code=404)

    def test_ajax_stats(self, model='driver', default_code=200):
        URL = 'dr27-ajax:stats'
        data = {'model': model}
        self._GET_request(URL, data=data, code=default_code)
        data['competition_slug'] = 'f1'
        self._GET_request(URL, data=data, code=default_code)
        data['year'] = 2016
        self._GET_request(URL, data=data, code=default_code)
        data['record'] = 'POLE'
        self._GET_request(URL, data=data, code=default_code)
        for opt in [None, 'streak', 'seasons', 'streak_top', 'streak_actives', 'streak_top_actives']:
            data['rank_opt'] = opt
            self._GET_request(URL, data=data, code=default_code)
        data['record'] = 'FFF'
        self._GET_request(URL, data=data, code=404)

    def test_ajax_stats_team(self):
        self.test_ajax_stats(model='team')

    def test_ajax_stats_404_model(self):
        self.test_ajax_stats(model='else', default_code=404)

    def test_race_view(self):
        kwargs = {'competition_slug': 'f1', 'year': 2016, 'race_id': 1}
        self._GET_request('season:race-view', kwargs=kwargs)
        kwargs['race_id'] = 200
        self._GET_request('season:race-view', kwargs=kwargs, code=404)


    def test_driver_records_global_view(self):
        kwargs = {}
        self._GET_request('global:index')
        self._GET_request('global:driver')
        self._GET_request('global:driver-olympic')
        self._GET_request('global:driver-seasons-rank')
        self._GET_request('global:driver-comeback')
        self._GET_request('global:driver-record-index')
        self._test_competition_view()
        self._test_season_view()

    def test_profiles_view(self):

        self._GET_request('global:driver-profile', kwargs={'driver_id': 1})
        self._GET_request('global:team-profile', kwargs={'team_id': 1})

        
    def test_team_records_global_view(self):
        self._GET_request('global:team')
        self._GET_request('global:team-olympic')
        self._GET_request('global:team-seasons-rank')
        self._GET_request('global:team-record-index')

    def test_contender_season_points(self):
        driver = Driver.objects.get(pk=1)
        season = Season.objects.get(pk=1)
        contender_season = ContenderSeason(driver=driver, season=season)

        punctuation_config_a = get_punctuation_config('F1-25')
        contender_season_points_a = contender_season.get_points(punctuation_config=punctuation_config_a)
        punctuation_config_b = get_punctuation_config('F1-10+6')
        contender_season_points_b = contender_season.get_points(punctuation_config=punctuation_config_b)
        self.assertGreater(contender_season_points_a, contender_season_points_b)

    def test_result_tuple(self):
        seat = Seat.objects.get(pk=1)
        results = seat.results.all()
        self.assertIsNotNone(get_tuples_from_results(results=results))

    def _check_get_changelist(self, ma):
        request = get_request()
        self.assertTrue(ma.get_changelist(request=request))

    def test_season_admin(self):
        ma = SeasonAdmin(Season, self.site)
        self._check_get_changelist(ma)
        self.assertTrue(ma.get_form(request=None, obj=None))
        self.assertIsInstance(ma.get_season_copy(copy_id=1), dict)

        request = get_request()
        season = Season.objects.get(pk=1)
        season_form = ma.get_form(request=request, obj=season)
        self.assertTrue(ma.print_copy_season(obj=season))
        self.assertIsNotNone(SeasonAdminForm(season_form))

    def _test_copy_url(self, COPY_URL, method_to_copy, data=None):
        season = Season.objects.get(pk=1)
        if data:
            request = self.factory.post(COPY_URL, data)
        else:
            request = self.factory.get(COPY_URL)
        ma = SeasonAdmin(Season, self.site)
        getattr(ma, method_to_copy)(request, season.pk)

    def _test_copy_items(self, COPY_URL, method_to_copy, new_season, items):
        self._test_copy_url(COPY_URL, method_to_copy)

        post_destiny = {
            'season_destiny': new_season.pk,
            'items': items,
            '_selector': True
        }

        self._test_copy_url(COPY_URL, method_to_copy, post_destiny)

        del post_destiny['_selector']
        post_destiny['_confirm'] = True

        self._test_copy_url(COPY_URL, method_to_copy, post_destiny)

    def test_season_copy_items(self):
        Season.objects.create(
            competition_id=1,
            year=2099,
            punctuation='F1-25',
            rounds=30
        )
        # todo check rounds
        new_season = Season.objects.get(competition_id=1, year=2099)

        season = Season.objects.get(pk=1)

        # races
        COPY_RACES_URL = reverse('admin:dr27-copy-races', kwargs={'pk': season.pk})
        races = [race.pk for race in season.races.all()]
        self._test_copy_items(COPY_RACES_URL, 'get_copy_races', new_season, races)

        ma = SeasonAdmin(Season, self.site)
        self.assertIn(COPY_RACES_URL, ma.print_copy_races(season))

    def test_season_duplicate(self):
        season = Season.objects.last()
        COPY_SEASON_URL = reverse('admin:dr27-copy-season', kwargs={'pk': season.pk})
        request = self.factory.get(COPY_SEASON_URL)
        ma = SeasonAdmin(Season, self.site)
        getattr(ma, 'get_duplicate_season')(request, season.pk)

        request = self.factory.post(COPY_SEASON_URL, data={'year': season.year, '_selector': True})
        ma = SeasonAdmin(Season, self.site)
        getattr(ma, 'get_duplicate_season')(request, season.pk)

        NEW_SEASON_YEAR = 9999
        request = self.factory.post(COPY_SEASON_URL, data={'year': NEW_SEASON_YEAR, '_selector': True})
        ma = SeasonAdmin(Season, self.site)
        getattr(ma, 'get_duplicate_season')(request, season.pk)

        self.assertTrue(Season.objects.get(competition=season.competition, year=NEW_SEASON_YEAR))

    def test_driver_admin(self):
        ma = DriverAdmin(Driver, self.site)
        self._check_get_changelist(ma)

    def test_team_admin(self):
        ma = TeamAdmin(Team, self.site)
        self._check_get_changelist(ma)
        team = Team.objects.get(pk=1)
        self.assertTrue(ma.print_competitions(team))

    def test_competition_admin(self):
        ma = CompetitionAdmin(Competition, self.site)
        self._check_get_changelist(ma)

    def test_circuit_admin(self):
        ma = CircuitAdmin(Circuit, self.site)
        self._check_get_changelist(ma)

    def test_grandprix_admin(self):
        ma = GrandPrixAdmin(GrandPrix, self.site)
        self._check_get_changelist(ma)

    def _test_season_formset(self, child_model, formset, fields):
        inline_formset = inlineformset_factory(Season, child_model, formset=formset,
                                               fields=fields,
                                               form=AlwaysChangedModelForm, can_delete=True)
        return inline_formset

    def _test_season_formset_copy(self, child_model, formset, fields, data=None):
        inline_formset = self._test_season_formset(child_model, formset, fields)
        # inline_formset.request = self.factory.request(QUERY_STRING='copy=1')
        related_formset = inline_formset(data)
        # self.assertTrue(related_formset.get_copy(copy_id=1))
        # self.assertFalse(related_formset.is_empty_form())
        self.assertFalse(related_formset.has_changed())
        return related_formset

    def test_race_formset(self):
        self._test_season_formset_copy(Race, RaceFormSet,
                                       ('round', 'grand_prix', 'circuit', 'date', 'alter_punctuation'))

    def test_race_admin(self):
        ma = RaceAdmin(Race, self.site)
        self._check_get_changelist(ma)
        race = Race.objects.get(pk=1)
        self.assertEquals(ma.print_pole(race), str(race.pole.driver))
        self.assertEquals(ma.print_winner(race), str(race.winner.driver))
        self.assertEquals(ma.print_fastest(race), str(race.fastest.driver))
        self.assertEquals(ma.clean_position('1'), 1)

    def _check_formfield_for_foreignkey(self, ma, request_obj, dbfield):
        request = get_request()
        request._obj_ = request_obj
        self.assertIsNotNone(ma.formfield_for_foreignkey(dbfield, request=request))

    def test_race_inline(self):
        race = Race.objects.get(pk=1)
        race_ma = RaceInline(SeasonAdmin, self.site)
        request = get_request()
        # self.assertIsNotNone(race_ma.get_formset())
        season = race.season
        self._check_formfield_for_foreignkey(race_ma, request_obj=season, dbfield=Race.grand_prix.field)

    # Currently not exists race with no results
    # def test_race_with_no_results(self):
    #     ma = RaceAdmin(Race, self.site)
    #     race = Race.objects.get(pk=20)  # No results
    #     self.assertIsNone(ma.print_pole(race))
    #     self.assertIsNone(ma.print_winner(race))
    #     self.assertIsNone(ma.print_fastest(race))

    def test_team_season_inline_admin(self):
        ma = TeamSeasonInline(SeasonAdmin, self.site)
        season = Season.objects.get(pk=1)
        self._check_formfield_for_foreignkey(ma, request_obj=season, dbfield=TeamSeason.team.field)
        team = Team.objects.get(pk=1)
        self._check_formfield_for_foreignkey(ma, request_obj=team, dbfield=TeamSeason.team.field)

    def test_related_competition_admin(self):
        race = Race.objects.get(pk=1)
        related_competition = RelatedCompetitionAdmin()
        # maybe exception
        self.assertIsNone(related_competition.print_competitions(race))


class DR27Api(APITestCase):
    fixtures = get_fixtures_test()

    def _GET_request(self, reverse_url, kwargs=None, code=200):
        reverse_url = ':'.join([DRIVER27_NAMESPACE, DRIVER27_API_NAMESPACE, reverse_url])
        request_url = reverse(reverse_url, kwargs=kwargs)
        response = self.client.get(request_url, format='json')
        self.assertEqual(response.status_code, code)

    def test_api_circuit(self):
        self._GET_request('circuit-list')

    def test_api_competition(self):
        self._GET_request('competition-list')
        self._GET_request('competition-detail', kwargs={'pk': 1})
        self._GET_request('competition-next-race', kwargs={'pk': 1})
        self._GET_request('competition-teams', kwargs={'pk': 1})

    def test_api_driver(self):
        self._GET_request('driver-list')
        self._GET_request('driver-detail', kwargs={'pk': 1})
        self._GET_request('driver-seats', kwargs={'pk': 1})

    def test_api_grand_prix(self):
        self._GET_request('grandprix-list')
        self._GET_request('grandprix-detail', kwargs={'pk': 1})

    def test_api_race(self):
        self._GET_request('race-list')
        self._GET_request('race-detail', kwargs={'pk': 1})
        self._GET_request('race-no-seats', kwargs={'pk': 1})
        self._GET_request('race-results', kwargs={'pk': 1})
        self._GET_request('race-seats', kwargs={'pk': 1})

    def test_api_result(self):
        self._GET_request('result-list')
        self._GET_request('result-detail', kwargs={'pk': 1})

    def test_api_season(self):
        self._GET_request('season-list')
        self._GET_request('season-detail', kwargs={'pk': 1})
        self._GET_request('season-next-race', kwargs={'pk': 1})
        self._GET_request('season-no-seats', kwargs={'pk': 1})
        self._GET_request('season-races', kwargs={'pk': 1})
        self._GET_request('season-seats', kwargs={'pk': 1})
        self._GET_request('season-standings', kwargs={'pk': 1})
        self._GET_request('season-standings-team', kwargs={'pk': 1})
        self._GET_request('season-teams', kwargs={'pk': 1})
        self._GET_request('season-title', kwargs={'pk': 1})

    def test_api_seat(self):
        self._GET_request('seat-list')
        self._GET_request('seat-detail', kwargs={'pk': 1})
        self._GET_request('seat-periods', kwargs={'pk': 1})
        self._GET_request('seatperiod-list')
        self._GET_request('seatperiod-detail', kwargs={'pk': 1})

    def test_api_team(self):
        self._GET_request('team-list')
        self._GET_request('team-detail', kwargs={'pk': 1})



from django import forms
from ..models import Race
from ..models import Seat
from ..models import TeamSeason


class RelatedWithSeasonFormSet(forms.models.BaseInlineFormSet):
    pass
    # model = None
    # model_attributes = []
    #
    # def get_copy(self, copy_id):
    #     related_objects = self.model.objects.filter(season__pk=copy_id)
    #     initial = []
    #     for x in related_objects:
    #         initial_x = {}
    #         for xa in self.model_attributes:
    #             initial_x[xa] = getattr(x, xa, None)
    #         initial.append(initial_x)
    #     return initial
    #
    # def __init__(self, *args, **kwargs):
    #     super(RelatedWithSeasonFormSet, self).__init__(*args, **kwargs)
    #     request = self.request
    #     copy_id = request.GET.get('copy')
    #     if not self.initial and copy_id:
    #         self.initial = self.get_copy(copy_id)
    #         self.extra += len(self.initial)


class RaceFormSet(RelatedWithSeasonFormSet):
    model = Race
    # model_attributes = ('round', 'circuit', 'grand_prix',)


class TeamSeasonFormSet(RelatedWithSeasonFormSet):
    model = TeamSeason



from django.conf.urls import url, include

from driver27 import views

urlpatterns = [
    url(r'^$', views.season_view, name='view'),
    url(r'^driver/position-draw/$', views.driver_season_pos_view, name='position-draw'),
    url(r'^driver/', include('driver27.urls.driver')),
    url(r'^team/', include('driver27.urls.team')),
    url(r'^race/', include('driver27.urls.season.race')),
]



from django.core.management.base import BaseCommand, CommandError
from driver27.models import Season
import csv


class Command(BaseCommand):
    help = 'Create CSV for import result from last season'

    def add_arguments(self, parser):
        parser.add_argument('csv',)

    def handle(self, *args, **options):
        season = Season.objects.order_by('year', 'competition__name').last()
        season_seats = season.seats.all()
        season_pending_races = season.pending_races.order_by('round')

        with open(options['csv'], 'wb') as csvfile:
            fieldnames = ['driver', 'team', 'seat_id', 'race_id', 'race_round', 'qualifying', 'finish', 'retired',
                          'fastest_lap', 'wildcard']
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()

            for race in season_pending_races:
                for seat in season_seats:
                    writer.writerow(
                        {
                            'driver': str(seat.driver),
                            'team': str(seat.team),
                            'seat_id': seat.pk,
                            'race_id': race.pk,
                            'race_round': race.round,
                            'qualifying': None,
                            'finish': None,
                            'retired': '',
                            'fastest_lap': '',
                            'wildcard': ''
                        }
                    )



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.core.exceptions import ValidationError
from .common import retro_encode, CommonRaceTestCase


class RaceTestCase(TestCase, CommonRaceTestCase):
    def test_circuit_unicode(self):
        circuit = self.get_test_circuit()
        expected_circuit = circuit.name
        self.assertEquals(str(circuit), retro_encode(expected_circuit))

    def test_race_unicode(self):
        race = self.get_test_race()
        expected_race = '{season}-{round}'.format(season=race.season, round=race.round)
        self.assertEquals(str(race), expected_race)
        return race

    def test_race_grandprix(self):
        race = self.get_test_race()
        season = race.season
        grand_prix = self.get_test_grandprix()
        # add grandprix without season
        race.grand_prix = grand_prix
        race.default_circuit = grand_prix.default_circuit
        # self.assertRaises(ValidationError, race.save)
        # # add season to grandprix
        # self.assertIsNone(grand_prix.competitions.add(season.competition))
        self.assertIsNone(race.save())
        # expected race changes (adding grandprix to str)
        expected_race = '{season}-{round}.{grand_prix}'.format(season=season, round=race.round,
                                                               grand_prix=grand_prix)
        self.assertEquals(str(race), expected_race)



"""
WSGI config for dr27app project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/howto/deployment/wsgi/
"""

import os
import sys

from django.core.wsgi import get_wsgi_application


def import_driver27():
    try:
        import driver27
    except ImportError:
        sys.path.append(os.path.abspath(os.path.join(os.path.abspath(__file__), '../../')))
        import driver27


import_driver27()

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "dr27app.settings")

application = get_wsgi_application()



from django.core.exceptions import ValidationError as DjangoValidationError
from rest_framework import viewsets, authentication, permissions, status, filters
from rest_framework.compat import set_rollback
from rest_framework.response import Response
from rest_framework.views import exception_handler as rest_exception_handler
from django_filters.rest_framework import DjangoFilterBackend
from collections import namedtuple



def get_dict_from_team_rank_entry(rank_entry):

    return {
        'points': rank_entry['points'],
        'team': rank_entry['team'].name,
        'positions_order': rank_entry['pos_str']
     }


def get_dict_from_races(races):
    return [
        {
            'id': race.pk,
            'name': u'{race}'.format(race=race),
            'alter_punctuation': race.alter_punctuation
        }
        for race in races
    ]


def get_dict_from_rank_entry(rank_entry):

    return {
        'points': rank_entry['points'],
        'driver': {
            'id': rank_entry['driver'].id,
            'last_name': rank_entry['driver'].last_name,
            'first_name': rank_entry['driver'].first_name,
            'country': rank_entry['driver'].country.code
        },
        'teams': rank_entry['teams'],
        'positions_order': rank_entry['pos_str']
     }


class DR27Serializer(object):
    def __init__(self, *args, **kwargs):
        self.exclude_fields = kwargs.pop('exclude_fields', None)
        super(DR27Serializer, self).__init__(*args, **kwargs)

    def get_field_names(self, declared_fields, info):
        fields = super(DR27Serializer, self).get_field_names(declared_fields, info)
        if getattr(self, 'exclude_fields', None):
            fields = tuple([x for x in fields if x not in self.exclude_fields])
        return fields


class DR27ViewSet(viewsets.ModelViewSet):
    # authentication_classes = (authentication.SessionAuthentication, authentication.BasicAuthentication)
    # permission_classes = (permissions.IsAuthenticated,)

    filter_backends = (DjangoFilterBackend, filters.SearchFilter,)

    def get_exception_handler(self):
        return custom_exception_handler


# based on:
# https://github.com/tomchristie/django-rest-framework/pull/3149/commits/413fe93dbfda41e3b7890ea9550d60bca3315761
# https://github.com/RockHoward/django-rest-framework
def custom_exception_handler(exc, context):
    response = rest_exception_handler(exc, context)
    if isinstance(exc, DjangoValidationError):
        data = {'detail': exc.messages}
        set_rollback()
        return Response(data, status=status.HTTP_400_BAD_REQUEST)
    return response


from .common import *
from .filters import *
from .formsets import *
from ..models import ContenderSeason
from ..models import Race
from ..models import Result
from ..models import Seat, SeatPeriod
from ..models import Team, TeamSeason


class RaceInlineForm(AlwaysChangedModelForm):
    class Meta:
        widgets = {
            'grand_prix': GrandPrixWidget,
        }


class RaceInline(CompetitionFilterInline):
    model = Race
    extra = 1
    # readonly_fields = ('print_results_link', )
    formset = RaceFormSet
    form = RaceInlineForm


class SeatInline(CompetitionFilterInline):
    model = Seat
    extra = 1



# class CompetitionTeamInline(admin.TabularInline):
#     model = Team.competitions.through
#     extra = 1


class TeamSeasonInlineForm(AlwaysChangedModelForm):
    pass


class TeamSeasonInline(CompetitionFilterInline):
    model = TeamSeason
    extra = 1
    formset = TeamSeasonFormSet
    form = TeamSeasonInlineForm


class TeamInline(admin.TabularInline):
    model = Team
    extra = 1


class ResultInline(CompetitionFilterInline):
    model = Result
    extra = 1
    ordering = ('retired', 'finish', 'qualifying',)
    readonly_fields = ('points',)

    @staticmethod
    def points(obj):
        points = obj.points
        return points if points else ''


class SeatPeriodInline(admin.TabularInline):
    model = SeatPeriod
    extra = 2



from .streak import Streak
from django.core.exceptions import ValidationError
from django.db import models
from .records import get_record_config
from . import LIMIT_POSITION_LIST


class AbstractStreakModel(object):
    def get_results(self, *args, **kwargs):
        raise NotImplementedError('Not implemented property')

    def get_reverse_results(self, *args, **kwargs):
        raise NotImplementedError('Not implemented property')

    @property
    def is_active(self):
        raise NotImplementedError('Not implemented property')

    def get_streak(self, max_streak=False, result_filter=None, unique_by_race=False, **filters):
        from .models import get_tuples_from_results
        if not result_filter:
            result_filter = {}
        results = self.get_reverse_results(**result_filter)
        results_tuples = get_tuples_from_results(results=results)
        return Streak(results=results_tuples, max_streak=max_streak, unique_by_race=unique_by_race).run(filters)

    def get_positions(self, qualifying=False, limit_races=None, competition=None, **kwargs):
        pos_field = 'qualifying' if qualifying else 'finish'
        results = self.get_results(limit_races=limit_races, competition=competition, **kwargs)
        positions = results.values_list(pos_field, flat=True)
        return list(positions)


    def get_positions_count_list(self, limit_races=None, competition=None, **kwargs):
        """ Return a list with the count of each 20 first positions """
        finish_pos = self.get_positions(limit_races=limit_races, competition=competition, **kwargs)
        positions_count = []
        for x in range(LIMIT_POSITION_LIST):
            position_count = finish_pos.count(x+1)
            positions_count.append(position_count)
        return positions_count

    def get_positions_count_str(self, position_list=None, limit_races=None):
        """ Return a str with position_count_list to order """
        " Each list item will be filled to zeros until get three digits e.g. 1 => 001, 12 => 012 "
        if not position_list:
            position_list = self.get_positions_count_list(limit_races=limit_races)
        positions_str = ''.join([str(x).zfill(3) for x in position_list])
        return positions_str


class AbstractStatsModel(models.Model):

    def get_results(self, *args, **kwargs):
        raise NotImplementedError('Not implemented property')

    @property
    def is_active(self):
        raise NotImplementedError('Not implemented property')


    def get_summary_points_rank(self, **kwargs):
        return self.get_summary_points(**kwargs)

    @property
    def result_filter_kwargs(self):
        raise NotImplementedError('Not implemented property')

    def season_stats_cls(self, *args, **kwargs):
        raise NotImplementedError('Not implemented method')

    def get_season(self, season):
        try:
            season_stats = self.season_stats_cls(season)
            return season_stats
        except ValidationError:
            return None

    def get_stats_list(self, records_list=None, append_points=False, **kwargs):
        multiple_records = {}
        if records_list is None:
            records_list = ['RACE', 'POLE', 'WIN', 'PODIUM', 'FASTEST']
        for record in records_list:
            record_config = get_record_config(record).get('filter')
            if hasattr(self, 'get_total_races') and record == 'RACE':
                record_stat = self.get_total_races(**dict(kwargs, **record_config))
            else:
                record_stat = self.get_stats(**dict(kwargs, **record_config))
            multiple_records[record] = record_stat
        if append_points:
            multiple_records['POINTS'] = self.get_points(**kwargs)
        return multiple_records

    def get_saved_points(self, limit_races=None, **kwargs):
        results = self.get_results(limit_races=limit_races, **kwargs)
        return [result.points for result in results if not result.wildcard and result.points is not None]

    def get_reverse_results(self, limit_races=None, **extra_filter):
        return self.get_results(limit_races=limit_races, reverse_order=True, **extra_filter)

    def get_races(self, **filters):
        """ Return only race id of team in season """
        results = self.get_results(**filters).values('race').annotate(count_race=models.Count('race')).order_by()
        return results

    def get_stats(self, **filters):
        """ Count 1 by each result """
        return self.get_results(**filters).count()

    class Meta:
        abstract = True


class TeamStatsModel(AbstractStatsModel, AbstractStreakModel):
    def season_stats_cls(self, *args, **kwargs):
        raise NotImplementedError('Not implemented method')


    @property
    def result_filter_kwargs(self):
        raise NotImplementedError('Not implemented property')

    @property
    def is_active(self):
        raise NotImplementedError('Not implemented property')

    def get_results(self, *args, **kwargs):
        raise NotImplementedError('Not implemented method')

    def get_total_races(self, **filters):
        """ Only count 1 by race with any driver in filter """
        return self.get_races(**filters).count()

    def get_total_stats(self, **filters):  # noqa
        """ Count 1 by each result """
        return self.get_results(**filters).count()

    def get_doubles_races(self, **filters):
        """ Only count 1 by race with at least two drivers in filter """
        return self.get_races(**filters).filter(count_race__gte=2).count()

    class Meta:
        abstract = True


class StatsByCompetitionModel(AbstractStatsModel, AbstractStreakModel):

    def get_points_by_season(self, season, **kwargs):
        raise NotImplementedError('Not implemented method')

    def get_summary_points(self, append_to_summary=None, **kwargs):
        raise NotImplementedError('Not implemented method')

    def get_stats_by_season(self, records_list=None, append_points=False, **kwargs):
        """
        Return multiple records (or only one) record in season

        """

        kwargs.pop('season', None)
        seasons = getattr(self, 'seasons').all()
        return [self.season_stats_cls(season=season) \
                    .get_summary_stats(records_list=records_list, append_points=append_points, **kwargs)
                for season in seasons]


    def get_stats_by_competition(self, records_list=None, append_points=False, **kwargs):
        competitions = getattr(self, 'competitions').all()
        return [
            {
                'competition': competition,
                'stats': self.get_stats_list(records_list=records_list,
                                             append_points=append_points,
                                             competition=competition, **kwargs)
            }
            for competition in competitions
        ]

    def seasons_by_competition(self, competition=None):
        seasons = getattr(self, 'seasons').all()
        if competition is not None:
            seasons = seasons.filter(competition=competition)
        return seasons

    def get_points(self, season=None, competition=None, punctuation_config=None):
        if season is not None:
            return self.get_season(season).get_points(punctuation_config=punctuation_config)
        seasons = self.seasons_by_competition(competition=competition)
        points = 0
        for season in seasons:
            season_points = self.get_season(season).get_points(punctuation_config=punctuation_config)
            points += season_points if season_points else 0
        return points

    class Meta:
        abstract = True

class SeasonStatsModel(object):

    def get_points_list(self, limit_races=None, punctuation_config=None, **kwargs):
        raise NotImplementedError('Not implemented method')

    def get_points(self, **kwargs):
        """ get_points in season must not be the same of other get_points because others is the sum of this """
        raise NotImplementedError('Not implemented method')

    def _summary_season(self, exclude_position=False):
        raise NotImplementedError('Not implemented method')

    def _points_position(self, rank, keyword):
        """
        Return position in season rank

        """
        attr = getattr(self, keyword)
        season = getattr(self, 'season')
        rank = getattr(season, rank)()
        position = None
        for index, entry in enumerate(rank):
            if attr == entry.get(keyword):
                position = index + 1
                break
        return position

    def get_summary_points_rank(self, **kwargs):
        return self.get_summary_points(points_to_rank=True, **kwargs)

    def get_summary_points(self, append_to_summary=None, **kwargs):
        exclude_position = kwargs.pop('exclude_position', False)
        punctuation_config = kwargs.pop('punctuation_config', None)
        summary_points = self._summary_season(exclude_position)
        if append_to_summary is None:
            append_to_summary = {}
        summary_points.update(
            points=self.get_points(punctuation_config=punctuation_config, **kwargs),
            pos_list=getattr(self, 'get_positions_count_list')(),
            pos_str=getattr(self, 'get_positions_count_str')(), **append_to_summary
        )


        return summary_points

    class Meta:
        abstract = True







from django.conf.urls import url, include

from .. import views_ajax as views
from ..api import urls_api

urlpatterns = [
    url(r'^standing/$', views.standing_view, name='standing'),
    url(r'^stats/$', views.stats_view, name='stats'),
]



from .models import Race, Team, Driver, Result


class Teammates(object):

    def group_results_by_race(self, results):
        group_by_race = {}
        for result in results:
            if result.race not in group_by_race:
                group_by_race[result.race] = []
            group_by_race[result.race].append(result)
        return group_by_race

    def group_results(self, results):
        results = results.order_by('race_id', 'seat__team')
        results_by_race = self.group_results_by_race(results)

    def wizard(self, race=None, team=None, driver=None):
        results = Result.wizard
        if race is not None:
            race_filter = {'race': race}
            results = results(**race_filter)
        if team is not None:
            team_filter = {'race': team.races.all()}
            results = results(**team_filter)
        if driver is not None:
            driver_filter = {'race': driver.races.all()}
            results = results(**driver_filter)
        self.group_results(results)

    class Meta:
        abstract = True



# -*- coding: utf-8 -*-
from django.http import Http404
from django.shortcuts import render
from django.utils.translation import ugettext as _

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .decorators import competition_request
from .records import get_record_config

def get_safe_record_config(record):
    record_config = get_record_config(record)
    if record_config:
        return record_config
    else:
        raise Http404(_('Record does not exist'))

@competition_request
def standing_view(request, context):
    season_or_competition = context.get('season_or_competition')

    # driver/team olympic/common
    standing_model = request.GET.get('model', 'driver')
    if standing_model not in ['driver', 'team']:
        raise Http404(_('The standing requested is invalid'))

    olympic = request.GET.get('olympic', False)
    if olympic:
        standing_type = suffix_tpl = 'olympic'
    else:
        standing_type = 'common'
        suffix_tpl = 'table'

    standing_method = {
        'driver': {'olympic': 'olympic_rank', 'common': 'points_rank'},
        'team': {'olympic': 'team_olympic_rank','common': 'team_points_rank'}
    }.get(standing_model)[standing_type]


    # kwargs to rank
    by_season = request.GET.get('by_season', False)
    punctuation_code = request.GET.get('punctuation', None)


    standing_kwargs = {}
    if not olympic:
        standing_kwargs.update(punctuation_code=punctuation_code, by_season=by_season)

    rank = getattr(season_or_competition, standing_method)(**standing_kwargs)
    tpl = 'driver27/' + standing_model + '/list-'+suffix_tpl + '.html'

    context.update(rank=rank, by_season=by_season)

    return render(request, tpl, context)

@competition_request
def stats_view(request, context):
    season_or_competition = context.get('season_or_competition')

    record = request.GET.get('record', None)
    record_config = get_safe_record_config(record)

    standing_model = request.GET.get('model', 'driver')

    if standing_model not in ['driver', 'team']:
        raise Http404(_('The standing requested is invalid'))

    rank_opt = request.GET.get('rank_opt', None)
    rank = None
    tpl = 'driver27/' + standing_model + '/record-table.html'
    if record:
        record_filter = record_config.get('filter')
        if rank_opt:
            if rank_opt == 'seasons':
                rank = stats_by_season_view(season_or_competition, standing_model, record_filter)
                context.update(rank_opt='seasons')
            else:
                rank = stats_by_streak_view(season_or_competition, standing_model, record_filter, rank_opt)
        else:
            rank = season_or_competition.stats_rank(**record_filter)

    context.update(rank=rank)
    return render(request, tpl, context)


def stats_by_streak_view(season_or_competition, standing_model, record_filter, rank_opt):
    streak_method = {
        'team': 'streak_team_rank',
        'driver': 'streak_rank'
    }.get(standing_model)
    rank_kwargs_by_opt = {
        'streak': {},
        'streak_top': {'max_streak': True},
        'streak_actives': {'only_actives': True},
        'streak_top_actives': {'only_actives': True, 'max_streak': True}
    }
    rank_kwargs = rank_kwargs_by_opt.get(rank_opt, {})
    record_filter.update(**rank_kwargs)
    rank = getattr(season_or_competition, streak_method)(**record_filter)
    return rank


def stats_by_season_view(season_or_competition, standing_model, record_filter):
    seasons_method = {
        'team': 'seasons_team_rank',
        'driver': 'seasons_rank'
    }.get(standing_model)
    rank = getattr(season_or_competition, seasons_method)(**record_filter)
    return rank


from django import template
from ..models import Season
register = template.Library()

@register.filter
def champion_filter(season_id):
    if season_id:
        season = Season.objects.get(pk=season_id)
        return '<span class="champion_tag">&#9818;</span>' if season.has_champion() else ''
    else:
        return ''


from rest_framework import routers
from .viewsets import CircuitViewSet, CompetitionViewSet, DriverViewSet
from .viewsets import GrandPrixViewSet, RaceViewSet, ResultViewSet, SeasonViewSet
from .viewsets import SeatViewSet, TeamViewSet, SeatPeriodViewSet
from ..common import DRIVER27_API_NAMESPACE

# Routers provide an easy way of automatically determining the URL conf.
router = routers.DefaultRouter()
router.register(r'circuits', CircuitViewSet)
router.register(r'competitions', CompetitionViewSet)
router.register(r'drivers', DriverViewSet)
router.register(r'grands-prix', GrandPrixViewSet)
router.register(r'races', RaceViewSet)
router.register(r'results', ResultViewSet)
router.register(r'seasons', SeasonViewSet)
router.register(r'seats', SeatViewSet)
router.register(r'seat-periods', SeatPeriodViewSet)
router.register(r'teams', TeamViewSet)

from django.conf.urls import url, include
urls_api = [
    url(r'^', include(router.urls, namespace=DRIVER27_API_NAMESPACE)),
]